@@ -115,6 +115,9 @@ static void gtk_expander_forall (GtkContainer *container,
 
 static void gtk_expander_activate (GtkExpander *expander);
 
+static void get_expander_bounds (GtkExpander  *expander,
+				 GdkRectangle *rect);
+
 static GtkBinClass *parent_class = NULL;
 
 GType
@@ -382,17 +385,20 @@ gtk_expander_realize (GtkWidget *widget)
   GdkWindowAttr attributes;
   gint attributes_mask;
   gint border_width;
+  GdkRectangle expander_rect;
 
   priv = GTK_EXPANDER (widget)->priv;
   GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
 
   border_width = GTK_CONTAINER (widget)->border_width;
-                                                                                                             
+
+  get_expander_bounds (GTK_EXPANDER (widget), &expander_rect);
+  
   attributes.window_type = GDK_WINDOW_CHILD;
   attributes.x = widget->allocation.x + border_width;
-  attributes.y = widget->allocation.y + border_width;
-  attributes.width = widget->allocation.width - 2 * border_width;
-  attributes.height = widget->allocation.height - 2 * border_width;
+  attributes.y = expander_rect.y;
+  attributes.width = MAX (widget->allocation.width - 2 * border_width, 1);
+  attributes.height = expander_rect.width;
   attributes.wclass = GDK_INPUT_ONLY;
   attributes.event_mask = gtk_widget_get_events (widget)     |
 				GDK_BUTTON_PRESS_MASK        |
@@ -608,9 +614,16 @@ gtk_expander_size_allocate (GtkWidget     *widget,
 
       ltr = gtk_widget_get_direction (widget) != GTK_TEXT_DIR_RTL;
 
-      label_allocation.x = widget->allocation.x + border_width + focus_width + focus_pad;
       if (ltr)
-	label_allocation.x += expander_size + 2 * expander_spacing;
+	label_allocation.x = (widget->allocation.x +
+                              border_width + focus_width + focus_pad +
+                              expander_size + 2 * expander_spacing);
+      else
+        label_allocation.x = (widget->allocation.x + widget->allocation.width -
+                              (label_requisition.width +
+                               border_width + focus_width + focus_pad +
+                               expander_size + 2 * expander_spacing));
+
       label_allocation.y = widget->allocation.y + border_width + focus_width + focus_pad;
 
       label_allocation.width = MIN (label_requisition.width,
@@ -699,6 +712,52 @@ gtk_expander_unmap (GtkWidget *widget)
     gtk_widget_unmap (priv->label_widget);
 }
 
+static void
+gtk_expander_paint_prelight (GtkExpander *expander)
+{
+  GtkWidget *widget;
+  GtkContainer *container;
+  GtkExpanderPrivate *priv;
+  GdkRectangle area;
+  gboolean interior_focus;
+  int focus_width;
+  int focus_pad;
+  int expander_size;
+  int expander_spacing;
+
+  priv = expander->priv;
+  widget = GTK_WIDGET (expander);
+  container = GTK_CONTAINER (expander);
+
+  gtk_widget_style_get (widget,
+			"interior-focus", &interior_focus,
+			"focus-line-width", &focus_width,
+			"focus-padding", &focus_pad,
+			"expander-size", &expander_size,
+			"expander-spacing", &expander_spacing,
+			NULL);
+
+  area.x = widget->allocation.x + container->border_width;
+  area.y = widget->allocation.y + container->border_width;
+  area.width = widget->allocation.width - (2 * container->border_width);
+
+  if (priv->label_widget && GTK_WIDGET_VISIBLE (priv->label_widget))
+    area.height = priv->label_widget->allocation.height;
+  else
+    area.height = 0;
+
+  area.height += interior_focus ? (focus_width + focus_pad) * 2 : 0;
+  area.height = MAX (area.height, expander_size + 2 * expander_spacing);
+  area.height += !interior_focus ? (focus_width + focus_pad) * 2 : 0;
+
+  gtk_paint_flat_box (widget->style, widget->window,
+		      GTK_STATE_PRELIGHT,
+		      GTK_SHADOW_ETCHED_OUT,
+		      &area, widget, "expander",
+		      area.x, area.y,
+		      area.width, area.height);
+}
+
 static void
 gtk_expander_paint (GtkExpander *expander)
 {
@@ -712,7 +771,11 @@ gtk_expander_paint (GtkExpander *expander)
 
   state = widget->state;
   if (expander->priv->prelight)
-    state = GTK_STATE_PRELIGHT;
+    {
+      state = GTK_STATE_PRELIGHT;
+
+      gtk_expander_paint_prelight (expander);
+    }
 
   gtk_paint_expander (widget->style,
 		      widget->window,
@@ -798,11 +861,6 @@ gtk_expander_expose (GtkWidget      *widget,
       if (GTK_WIDGET_HAS_FOCUS (expander))
 	gtk_expander_paint_focus (expander, &event->area);
 
-      if (expander->priv->label_widget)
-	gtk_container_propagate_expose (GTK_CONTAINER (widget),
-					expander->priv->label_widget,
-					event);
-
       GTK_WIDGET_CLASS (parent_class)->expose_event (widget, event);
     }
 
@@ -859,15 +917,12 @@ gtk_expander_state_changed (GtkWidget    *widget,
 static void
 gtk_expander_redraw_expander (GtkExpander *expander)
 {
-  GdkRectangle bounds;
+  GtkWidget *widget;
 
-  get_expander_bounds (expander, &bounds);
+  widget = GTK_WIDGET (expander);
 
-  gtk_widget_queue_draw_area (GTK_WIDGET (expander),
-			      bounds.x,
-			      bounds.y,
-			      bounds.width,
-			      bounds.height);
+  if (GTK_WIDGET_REALIZED (widget))
+    gdk_window_invalidate_rect (widget->window, &widget->allocation, FALSE);
 }
 
 static gboolean
@@ -883,6 +938,10 @@ gtk_expander_enter_notify (GtkWidget        *widget,
       event->detail != GDK_NOTIFY_INFERIOR)
     {
       expander->priv->prelight = TRUE;
+
+      if (expander->priv->label_widget)
+	gtk_widget_set_state (expander->priv->label_widget, GTK_STATE_PRELIGHT);
+
       gtk_expander_redraw_expander (expander);
     }
 
@@ -902,6 +961,10 @@ gtk_expander_leave_notify (GtkWidget        *widget,
       event->detail != GDK_NOTIFY_INFERIOR)
     {
       expander->priv->prelight = FALSE;
+
+      if (expander->priv->label_widget)
+	gtk_widget_set_state (expander->priv->label_widget, GTK_STATE_NORMAL);
+
       gtk_expander_redraw_expander (expander);
     }
 
@@ -1532,14 +1595,21 @@ gtk_expander_set_label_widget (GtkExpander *expander,
     return;
 
   if (priv->label_widget)
-    gtk_widget_unparent (priv->label_widget);
+    {
+      gtk_widget_set_state (priv->label_widget, GTK_STATE_NORMAL);
+      gtk_widget_unparent (priv->label_widget);
+    }
 
   priv->label_widget = label_widget;
 
   if (label_widget)
     {
       priv->label_widget = label_widget;
+
       gtk_widget_set_parent (label_widget, GTK_WIDGET (expander));
+
+      if (priv->prelight)
+	gtk_widget_set_state (label_widget, GTK_STATE_PRELIGHT);
     }
 
   if (GTK_WIDGET_VISIBLE (expander))
