@@ -32,12 +32,16 @@
 #include <string.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <pwd.h>
 #include <unistd.h>
 #include <stdio.h>
+#include <time.h>
 
 #define BOOKMARKS_FILENAME ".gtk-bookmarks"
 #define BOOKMARKS_TMP_FILENAME ".gtk-bookmarks-XXXXXX"
 
+#define FOLDER_CACHE_LIFETIME 2 /* seconds */
+
 typedef struct _GtkFileSystemUnixClass GtkFileSystemUnixClass;
 
 #define GTK_FILE_SYSTEM_UNIX_CLASS(klass)     (G_TYPE_CHECK_CLASS_CAST ((klass), GTK_TYPE_FILE_SYSTEM_UNIX, GtkFileSystemUnixClass))
@@ -59,6 +63,7 @@ struct _GtkFileSystemUnix
 /* Icon type, supplemented by MIME type
  */
 typedef enum {
+  ICON_UNDECIDED,
   ICON_NONE,
   ICON_REGULAR,	/* Use mime type for icon */
   ICON_BLOCK_DEVICE,
@@ -93,8 +98,23 @@ struct _GtkFileFolderUnix
   GtkFileSystemUnix *system_unix;
   GtkFileInfoType types;
   gchar *filename;
+  GHashTable *stat_info;
+  unsigned int have_stat : 1;
+  unsigned int have_mime_type : 1;
+  time_t asof;
 };
 
+struct stat_info_entry {
+  struct stat statbuf;
+  char *mime_type;
+  IconType icon_type;
+};
+
+static const GtkFileInfoType STAT_NEEDED_MASK = (GTK_FILE_INFO_IS_FOLDER |
+						 GTK_FILE_INFO_IS_HIDDEN |
+						 GTK_FILE_INFO_MODIFICATION_TIME |
+						 GTK_FILE_INFO_SIZE);
+
 static GObjectClass *system_parent_class;
 static GObjectClass *folder_parent_class;
 
@@ -184,12 +204,17 @@ static gboolean     gtk_file_folder_unix_list_children (GtkFileFolder  *folder,
 							GSList        **children,
 							GError        **error);
 
+static gboolean     gtk_file_folder_unix_is_finished_loading (GtkFileFolder *folder);
+
 static GtkFilePath *filename_to_path   (const gchar       *filename);
 
 static gboolean     filename_is_root  (const char       *filename);
-static GtkFileInfo *filename_get_info (const gchar      *filename,
-				       GtkFileInfoType   types,
-				       GError          **error);
+
+static gboolean fill_in_names (GtkFileFolderUnix *folder_unix, GError **error);
+static gboolean fill_in_stats (GtkFileFolderUnix *folder_unix, GError **error);
+static gboolean fill_in_mime_type (GtkFileFolderUnix *folder_unix, GError **error);
+
+static char *       get_parent_dir    (const char       *filename);
 
 /*
  * GtkFileSystemUnix
@@ -322,6 +347,19 @@ gtk_file_system_unix_get_volume_for_path (GtkFileSystem     *file_system,
   return get_root_volume ();
 }
 
+static char *
+remove_trailing_slash (const char *filename)
+{
+  int len;
+
+  len = strlen (filename);
+
+  if (len > 1 && filename[len - 1] == '/')
+    return g_strndup (filename, len - 1);
+  else
+    return g_memdup (filename, len + 1);
+}
+
 static GtkFileFolder *
 gtk_file_system_unix_get_folder (GtkFileSystem     *file_system,
 				 const GtkFilePath *path,
@@ -331,6 +369,8 @@ gtk_file_system_unix_get_folder (GtkFileSystem     *file_system,
   GtkFileSystemUnix *system_unix;
   GtkFileFolderUnix *folder_unix;
   const char *filename;
+  char *filename_copy;
+  time_t now = time (NULL);
 
   system_unix = GTK_FILE_SYSTEM_UNIX (file_system);
 
@@ -338,24 +378,85 @@ gtk_file_system_unix_get_folder (GtkFileSystem     *file_system,
   g_return_val_if_fail (filename != NULL, NULL);
   g_return_val_if_fail (g_path_is_absolute (filename), NULL);
 
-  folder_unix = g_hash_table_lookup (system_unix->folder_hash, filename);
+  filename_copy = remove_trailing_slash (filename);
+  folder_unix = g_hash_table_lookup (system_unix->folder_hash, filename_copy);
 
   if (folder_unix)
     {
+      g_free (filename_copy);
+      if (now - folder_unix->asof >= FOLDER_CACHE_LIFETIME &&
+	  folder_unix->stat_info)
+	{
+#if 0
+	  g_print ("Cleaning out cached directory %s\n", filename);
+#endif
+	  g_hash_table_destroy (folder_unix->stat_info);
+	  folder_unix->stat_info = NULL;
+	  folder_unix->have_mime_type = FALSE;
+	  folder_unix->have_stat = FALSE;
+	}
+
+      g_object_ref (folder_unix);
       folder_unix->types |= types;
-      return g_object_ref (folder_unix);
+      types = folder_unix->types;
     }
   else
     {
+      if (!g_file_test (filename, G_FILE_TEST_IS_DIR))
+	{
+	  int save_errno = errno;
+	  gchar *filename_utf8 = g_filename_to_utf8 (filename, -1, NULL, NULL, NULL);
+
+	  /* If g_file_test() returned FALSE but not due to an error, it means
+	   * that the filename is not a directory.
+	   */
+	  if (save_errno == 0)
+	    /* ENOTDIR */
+	    g_set_error (error,
+			 GTK_FILE_SYSTEM_ERROR,
+			 GTK_FILE_SYSTEM_ERROR_NOT_FOLDER,
+			 _("%s: %s"),
+			 filename_utf8 ? filename_utf8 : "???",
+			 g_strerror (ENOTDIR));
+	  else
+	    g_set_error (error,
+			 GTK_FILE_SYSTEM_ERROR,
+			 GTK_FILE_SYSTEM_ERROR_NONEXISTENT,
+			 _("error getting information for '%s': %s"),
+			 filename_utf8 ? filename_utf8 : "???",
+			 g_strerror (save_errno));
+
+	  g_free (filename_utf8);
+	  g_free (filename_copy);
+	  return NULL;
+	}
+
       folder_unix = g_object_new (GTK_TYPE_FILE_FOLDER_UNIX, NULL);
       folder_unix->system_unix = system_unix;
-      folder_unix->filename = g_strdup (filename);
+      folder_unix->filename = filename_copy;
       folder_unix->types = types;
+      folder_unix->stat_info = NULL;
+      folder_unix->asof = now;
+      folder_unix->have_mime_type = FALSE;
+      folder_unix->have_stat = FALSE;
 
-      g_hash_table_insert (system_unix->folder_hash, folder_unix->filename, folder_unix);
-
-      return GTK_FILE_FOLDER (folder_unix);
+      g_hash_table_insert (system_unix->folder_hash,
+			   folder_unix->filename,
+			   folder_unix);
     }
+
+  if ((types & STAT_NEEDED_MASK) && !fill_in_stats (folder_unix, error))
+    {
+      g_object_unref (folder_unix);
+      return NULL;
+    }
+  if ((types & GTK_FILE_INFO_MIME_TYPE) && !fill_in_mime_type (folder_unix, error))
+    {
+      g_object_unref (folder_unix);
+      return NULL;
+    }
+
+  return GTK_FILE_FOLDER (folder_unix);
 }
 
 static gboolean
@@ -366,7 +467,7 @@ gtk_file_system_unix_create_folder (GtkFileSystem     *file_system,
   GtkFileSystemUnix *system_unix;
   const char *filename;
   gboolean result;
-  char *parent;
+  char *parent, *tmp;
 
   system_unix = GTK_FILE_SYSTEM_UNIX (file_system);
 
@@ -374,7 +475,9 @@ gtk_file_system_unix_create_folder (GtkFileSystem     *file_system,
   g_return_val_if_fail (filename != NULL, FALSE);
   g_return_val_if_fail (g_path_is_absolute (filename), FALSE);
 
-  result = mkdir (filename, 0777) == 0;
+  tmp = remove_trailing_slash (filename);
+  result = mkdir (tmp, 0777) == 0;
+  g_free (tmp);
 
   if (!result)
     {
@@ -393,7 +496,7 @@ gtk_file_system_unix_create_folder (GtkFileSystem     *file_system,
   if (filename_is_root (filename))
     return TRUE; /* hmmm, but with no notification */
 
-  parent = g_path_get_dirname (filename);
+  parent = get_parent_dir (filename);
   if (parent)
     {
       GtkFileFolderUnix *folder_unix;
@@ -401,12 +504,31 @@ gtk_file_system_unix_create_folder (GtkFileSystem     *file_system,
       folder_unix = g_hash_table_lookup (system_unix->folder_hash, parent);
       if (folder_unix)
 	{
+	  GtkFileInfoType types;
+	  GtkFilePath *parent_path;
 	  GSList *paths;
+	  GtkFileFolder *folder;
 
-	  paths = g_slist_append (NULL, (GtkFilePath *) path);
-	  g_signal_emit_by_name (folder_unix, "files-added", paths);
-	  g_slist_free (paths);
+	  /* This is sort of a hack.  We re-get the folder, to ensure that the
+	   * newly-created directory gets read into the folder's info hash table.
+	   */
+
+	  types = folder_unix->types;
+
+	  parent_path = gtk_file_path_new_dup (parent);
+	  folder = gtk_file_system_get_folder (file_system, parent_path, types, NULL);
+	  gtk_file_path_free (parent_path);
+
+	  if (folder)
+	    {
+	      paths = g_slist_append (NULL, (GtkFilePath *) path);
+	      g_signal_emit_by_name (folder, "files-added", paths);
+	      g_slist_free (paths);
+	      g_object_unref (folder);
+	    }
 	}
+
+      g_free (parent);
     }
 
   return TRUE;
@@ -455,12 +577,30 @@ gtk_file_system_unix_volume_get_display_name (GtkFileSystem       *file_system,
   return g_strdup (_("Filesystem")); /* Same as Nautilus */
 }
 
+static IconType
+get_icon_type_from_stat (struct stat *statp)
+{
+  if (S_ISBLK (statp->st_mode))
+    return ICON_BLOCK_DEVICE;
+  else if (S_ISLNK (statp->st_mode))
+    return ICON_BROKEN_SYMBOLIC_LINK; /* See get_icon_type */
+  else if (S_ISCHR (statp->st_mode))
+    return ICON_CHARACTER_DEVICE;
+  else if (S_ISDIR (statp->st_mode))
+    return ICON_DIRECTORY;
+  else if (S_ISFIFO (statp->st_mode))
+    return  ICON_FIFO;
+  else if (S_ISSOCK (statp->st_mode))
+    return ICON_SOCKET;
+  else
+    return ICON_REGULAR;
+}
+
 static IconType
 get_icon_type (const char *filename,
 	       GError    **error)
 {
   struct stat statbuf;
-  IconType icon_type;
 
   /* If stat fails, try to fall back to lstat to catch broken links
    */
@@ -482,33 +622,7 @@ get_icon_type (const char *filename,
 	}
     }
 
-  if (S_ISBLK (statbuf.st_mode))
-    icon_type = ICON_BLOCK_DEVICE;
-  else if (S_ISLNK (statbuf.st_mode))
-    icon_type = ICON_BROKEN_SYMBOLIC_LINK; /* See above */
-  else if (S_ISCHR (statbuf.st_mode))
-    icon_type = ICON_CHARACTER_DEVICE;
-  else if (S_ISDIR (statbuf.st_mode))
-    icon_type = ICON_DIRECTORY;
-  else if (S_ISFIFO (statbuf.st_mode))
-    icon_type =  ICON_FIFO;
-  else if (S_ISSOCK (statbuf.st_mode))
-    icon_type = ICON_SOCKET;
-  else
-    {
-      icon_type = ICON_REGULAR;
-
-#if 0
-      if ((types & GTK_FILE_INFO_ICON) && icon_type == GTK_FILE_ICON_REGULAR &&
-	  (statbuf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)) &&
-	  (strcmp (mime_type, XDG_MIME_TYPE_UNKNOWN) == 0 ||
-	   strcmp (mime_type, "application/x-executable") == 0 ||
-	   strcmp (mime_type, "application/x-shellscript") == 0))
-	gtk_file_info_set_icon_type (info, GTK_FILE_ICON_EXECUTABLE);
-#endif
-    }
-
-  return icon_type;
+  return get_icon_type_from_stat (&statbuf);
 }
 
 typedef struct
@@ -591,6 +705,29 @@ gtk_file_system_unix_volume_render_icon (GtkFileSystem        *file_system,
   return get_cached_icon (widget, "gnome-fs-blockdev", pixel_size);
 }
 
+static char *
+get_parent_dir (const char *filename)
+{
+  int len;
+
+  len = strlen (filename);
+
+  /* Ignore trailing slashes */
+  if (len > 1 && filename[len - 1] == '/')
+    {
+      char *tmp, *parent;
+
+      tmp = g_strndup (filename, len - 1);
+
+      parent = g_path_get_dirname (tmp);
+      g_free (tmp);
+
+      return parent;
+    }
+  else
+    return g_path_get_dirname (filename);
+}
+
 static gboolean
 gtk_file_system_unix_get_parent (GtkFileSystem     *file_system,
 				 const GtkFilePath *path,
@@ -609,7 +746,7 @@ gtk_file_system_unix_get_parent (GtkFileSystem     *file_system,
     }
   else
     {
-      gchar *parent_filename = g_path_get_dirname (filename);
+      gchar *parent_filename = get_parent_dir (filename);
       *parent = filename_to_path (parent_filename);
       g_free (parent_filename);
     }
@@ -633,6 +770,18 @@ gtk_file_system_unix_make_path (GtkFileSystem    *file_system,
   g_return_val_if_fail (base_filename != NULL, NULL);
   g_return_val_if_fail (g_path_is_absolute (base_filename), NULL);
 
+  if (strchr (display_name, G_DIR_SEPARATOR))
+    {
+      g_set_error (error,
+		   GTK_FILE_SYSTEM_ERROR,
+		   GTK_FILE_SYSTEM_ERROR_BAD_FILENAME,
+		   _("The name \"%s\" is not valid because it contains the character \"%s\". "
+		     "Please use a different name."),
+		   display_name,
+		   G_DIR_SEPARATOR_S);
+      return NULL;
+    }
+
   filename = g_filename_from_utf8 (display_name, -1, NULL, NULL, &tmp_error);
   if (!filename)
     {
@@ -728,6 +877,48 @@ canonicalize_filename (gchar *filename)
   *q = '\0';
 }
 
+/* Takes a user-typed filename and expands a tilde at the beginning of the string */
+static char *
+expand_tilde (const char *filename)
+{
+  const char *notilde;
+  const char *slash;
+  const char *home;
+
+  if (filename[0] != '~')
+    return g_strdup (filename);
+
+  notilde = filename + 1;
+
+  slash = strchr (notilde, G_DIR_SEPARATOR);
+  if (!slash)
+    return NULL;
+
+  if (slash == notilde)
+    {
+      home = g_get_home_dir ();
+
+      if (!home)
+	return g_strdup (filename);
+    }
+  else
+    {
+      char *username;
+      struct passwd *passwd;
+
+      username = g_strndup (notilde, slash - notilde);
+      passwd = getpwnam (username);
+      g_free (username);
+
+      if (!passwd)
+	return g_strdup (filename);
+
+      home = passwd->pw_dir;
+    }
+
+  return g_build_filename (home, G_DIR_SEPARATOR_S, slash + 1, NULL);
+}
+
 static gboolean
 gtk_file_system_unix_parse (GtkFileSystem     *file_system,
 			    const GtkFilePath *base_path,
@@ -737,6 +928,7 @@ gtk_file_system_unix_parse (GtkFileSystem     *file_system,
 			    GError           **error)
 {
   const char *base_filename;
+  gchar *filename;
   gchar *last_slash;
   gboolean result = FALSE;
 
@@ -744,11 +936,21 @@ gtk_file_system_unix_parse (GtkFileSystem     *file_system,
   g_return_val_if_fail (base_filename != NULL, FALSE);
   g_return_val_if_fail (g_path_is_absolute (base_filename), FALSE);
 
-  last_slash = strrchr (str, G_DIR_SEPARATOR);
+  filename = expand_tilde (str);
+  if (!filename)
+    {
+      g_set_error (error,
+		   GTK_FILE_SYSTEM_ERROR,
+		   GTK_FILE_SYSTEM_ERROR_BAD_FILENAME,
+		   "%s", ""); /* nothing for now, as we are string-frozen */
+      return FALSE;
+    }
+
+  last_slash = strrchr (filename, G_DIR_SEPARATOR);
   if (!last_slash)
     {
       *folder = gtk_file_path_copy (base_path);
-      *file_part = g_strdup (str);
+      *file_part = g_strdup (filename);
       result = TRUE;
     }
   else
@@ -757,10 +959,10 @@ gtk_file_system_unix_parse (GtkFileSystem     *file_system,
       gchar *folder_path;
       GError *tmp_error = NULL;
 
-      if (last_slash == str)
+      if (last_slash == filename)
 	folder_part = g_strdup ("/");
       else
-	folder_part = g_filename_from_utf8 (str, last_slash - str,
+	folder_part = g_filename_from_utf8 (filename, last_slash - filename,
 					    NULL, NULL, &tmp_error);
 
       if (!folder_part)
@@ -793,6 +995,8 @@ gtk_file_system_unix_parse (GtkFileSystem     *file_system,
 	}
     }
 
+  g_free (filename);
+
   return result;
 }
 
@@ -856,10 +1060,53 @@ gtk_file_system_unix_render_icon (GtkFileSystem     *file_system,
 {
   const char *filename;
   IconType icon_type;
-  const char *mime_type;
+  const char *mime_type = NULL;
+  char *dirname;
+  GtkFileSystemUnix *system_unix;
+  GtkFileFolderUnix *folder_unix;
 
+  system_unix = GTK_FILE_SYSTEM_UNIX (file_system);
   filename = gtk_file_path_get_string (path);
-  icon_type = get_icon_type (filename, error);
+  dirname = g_path_get_dirname (filename);
+  folder_unix = g_hash_table_lookup (system_unix->folder_hash, dirname);
+  g_free (dirname);
+
+  if (folder_unix)
+    {
+      char *basename;
+      struct stat_info_entry *entry;
+
+      if (!fill_in_stats (folder_unix, error))
+	return NULL;
+
+      basename = g_path_get_basename (filename);
+      entry = g_hash_table_lookup (folder_unix->stat_info, basename);
+      g_free (basename);
+      if (entry)
+	{
+	  if (entry->icon_type == ICON_UNDECIDED)
+	    entry->icon_type = get_icon_type_from_stat (&entry->statbuf);
+	  icon_type = entry->icon_type;
+	  if (icon_type == ICON_REGULAR)
+	    {
+	      (void)fill_in_mime_type (folder_unix, NULL);
+	      mime_type = entry->mime_type;
+	    }
+	}
+      else
+	icon_type = ICON_NONE;
+    }
+  else
+    {
+#if 0
+      g_print ("No folder open for %s\n", filename);
+#endif
+
+      icon_type = get_icon_type (filename, error);
+      if (icon_type == ICON_REGULAR)
+	mime_type = xdg_mime_get_mime_type_for_file (filename);
+    }
+
 
   /* FIXME: this function should not return NULL without setting the GError; we
    * should perhaps provide a "never fails" generic stock icon for when all else
@@ -904,7 +1151,6 @@ gtk_file_system_unix_render_icon (GtkFileSystem     *file_system,
       return get_cached_icon (widget, name, pixel_size);
     }
 
-  mime_type = xdg_mime_get_mime_type_for_file (filename);
   if (mime_type)
     {
       const char *separator;
@@ -1288,6 +1534,7 @@ gtk_file_folder_unix_iface_init (GtkFileFolderIface *iface)
 {
   iface->get_info = gtk_file_folder_unix_get_info;
   iface->list_children = gtk_file_folder_unix_list_children;
+  iface->is_finished_loading = gtk_file_folder_unix_is_finished_loading;
 }
 
 static void
@@ -1302,166 +1549,285 @@ gtk_file_folder_unix_finalize (GObject *object)
 
   g_hash_table_remove (folder_unix->system_unix->folder_hash, folder_unix->filename);
 
+  if (folder_unix->stat_info)
+    {
+#if 0
+      g_print ("Releasing information for directory %s\n", folder_unix->filename);
+#endif
+      g_hash_table_destroy (folder_unix->stat_info);
+    }
+
   g_free (folder_unix->filename);
 
   folder_parent_class->finalize (object);
 }
 
 static GtkFileInfo *
-gtk_file_folder_unix_get_info (GtkFileFolder  *folder,
-			       const GtkFilePath    *path,
-			       GError        **error)
+gtk_file_folder_unix_get_info (GtkFileFolder      *folder,
+			       const GtkFilePath  *path,
+			       GError            **error)
 {
   GtkFileFolderUnix *folder_unix = GTK_FILE_FOLDER_UNIX (folder);
   GtkFileInfo *info;
-  gchar *dirname;
+  gchar *dirname, *basename;
   const char *filename;
+  struct stat_info_entry *entry;
+  gboolean file_must_exist;
+  GtkFileInfoType types;
+
+  if (!path)
+    {
+      struct stat buf;
+
+      g_return_val_if_fail (filename_is_root (folder_unix->filename), NULL);
+
+      if (stat (folder_unix->filename, &buf) != 0)
+	return NULL;
+
+      info = gtk_file_info_new ();
+      gtk_file_info_set_display_name (info, "/");
+      gtk_file_info_set_is_folder (info, TRUE);
+      gtk_file_info_set_is_hidden (info, FALSE);
+      gtk_file_info_set_mime_type (info, "x-directory/normal");
+      gtk_file_info_set_modification_time (info, buf.st_mtime);
+      gtk_file_info_set_size (info, buf.st_size);
+
+      return info;
+    }
 
   filename = gtk_file_path_get_string (path);
   g_return_val_if_fail (filename != NULL, NULL);
   g_return_val_if_fail (g_path_is_absolute (filename), NULL);
 
-  dirname = g_path_get_dirname (filename);
+  dirname = get_parent_dir (filename);
   g_return_val_if_fail (strcmp (dirname, folder_unix->filename) == 0, NULL);
   g_free (dirname);
 
-  info = filename_get_info (filename, folder_unix->types, error);
+  basename = g_path_get_basename (filename);
+  types = folder_unix->types;
+  file_must_exist = (types & ~GTK_FILE_INFO_DISPLAY_NAME) != 0;
+  entry = file_must_exist
+    ? g_hash_table_lookup (folder_unix->stat_info, basename)
+    : NULL;
+  /* basename freed later.  */
+
+  if (!file_must_exist || entry)
+    {
+      info = gtk_file_info_new ();
+    }
+  else
+    {
+      gchar *filename_utf8 = g_filename_to_utf8 (filename, -1, NULL, NULL, NULL);
+      g_set_error (error,
+		   GTK_FILE_SYSTEM_ERROR,
+		   GTK_FILE_SYSTEM_ERROR_NONEXISTENT,
+		   _("error getting information for '%s'"),
+		   filename_utf8 ? filename_utf8 : "???");
+      g_free (filename_utf8);
+      info = NULL;
+      types = 0;
+    }
+
+  if (types & GTK_FILE_INFO_DISPLAY_NAME)
+    {
+      gchar *display_name = g_filename_to_utf8 (basename, -1, NULL, NULL, NULL);
+      if (!display_name)
+	display_name = g_strescape (basename, NULL);
+
+      gtk_file_info_set_display_name (info, display_name);
+
+      g_free (display_name);
+    }
+
+  if (types & GTK_FILE_INFO_IS_HIDDEN)
+    gtk_file_info_set_is_hidden (info, basename[0] == '.');
+
+  if (types & GTK_FILE_INFO_IS_FOLDER)
+    gtk_file_info_set_is_folder (info, S_ISDIR (entry->statbuf.st_mode));
+
+  if (types & GTK_FILE_INFO_MIME_TYPE)
+    gtk_file_info_set_mime_type (info, entry->mime_type);
+
+  if (types & GTK_FILE_INFO_MODIFICATION_TIME)
+    gtk_file_info_set_modification_time (info, entry->statbuf.st_mtime);
+
+  if (types & GTK_FILE_INFO_SIZE)
+    gtk_file_info_set_size (info, (gint64)entry->statbuf.st_size);
+
+  g_free (basename);
 
   return info;
 }
 
+
+static void
+cb_list_children (gpointer key, gpointer value, gpointer user_data)
+{
+  GSList **children = user_data;
+  *children = g_slist_prepend (*children, key);
+}
+
 static gboolean
 gtk_file_folder_unix_list_children (GtkFileFolder  *folder,
 				    GSList        **children,
 				    GError        **error)
 {
   GtkFileFolderUnix *folder_unix = GTK_FILE_FOLDER_UNIX (folder);
-  GError *tmp_error = NULL;
-  GDir *dir;
+  GSList *l;
+
+  if (!fill_in_names (folder_unix, error))
+    return FALSE;
 
   *children = NULL;
 
-  dir = g_dir_open (folder_unix->filename, 0, &tmp_error);
+  /* Get the list of basenames.  */
+  g_hash_table_foreach (folder_unix->stat_info, cb_list_children, children);
+
+  /* Turn basenames into GFilePaths.  */
+  for (l = *children; l; l = l->next)
+    {
+      const char *basename = l->data;
+      char *fullname = g_build_filename (folder_unix->filename, basename, NULL);
+      l->data = filename_to_path (fullname);
+      g_free (fullname);
+    }
+  return TRUE;
+}
+
+static gboolean
+gtk_file_folder_unix_is_finished_loading (GtkFileFolder *folder)
+{
+  /* Since we don't do asynchronous loads, we are always finished loading */
+  return TRUE;
+}
+
+static void
+free_stat_info_entry (struct stat_info_entry *entry)
+{
+  g_free (entry->mime_type);
+  g_free (entry);
+}
+
+static gboolean
+fill_in_names (GtkFileFolderUnix *folder_unix, GError **error)
+{
+  GDir *dir;
+
+  if (folder_unix->stat_info)
+    return TRUE;
+
+#if 0
+  g_print ("Reading directory %s\n", folder_unix->filename);
+#endif
+
+  folder_unix->stat_info = g_hash_table_new_full (g_str_hash, g_str_equal,
+						  (GDestroyNotify)g_free,
+						  (GDestroyNotify)free_stat_info_entry);
+  dir = g_dir_open (folder_unix->filename, 0, error);
   if (!dir)
     {
+      int save_errno = errno;
+      gchar *filename_utf8 = g_filename_to_utf8 (folder_unix->filename, -1, NULL, NULL, NULL);
       g_set_error (error,
 		   GTK_FILE_SYSTEM_ERROR,
 		   GTK_FILE_SYSTEM_ERROR_NONEXISTENT,
-		   "%s",
-		   tmp_error->message);
-
-      g_error_free (tmp_error);
-
+		   _("error getting information for '%s': %s"),
+		   filename_utf8 ? filename_utf8 : "???",
+		   g_strerror (save_errno));
+      g_free (filename_utf8);
       return FALSE;
     }
 
   while (TRUE)
     {
-      const gchar *filename = g_dir_read_name (dir);
-      gchar *fullname;
-
-      if (!filename)
+      const gchar *basename = g_dir_read_name (dir);
+      if (!basename)
 	break;
 
-      fullname = g_build_filename (folder_unix->filename, filename, NULL);
-      *children = g_slist_prepend (*children, filename_to_path (fullname));
-      g_free (fullname);
+      g_hash_table_insert (folder_unix->stat_info,
+			   g_strdup (basename),
+			   g_new0 (struct stat_info_entry, 1));
     }
 
   g_dir_close (dir);
 
-  *children = g_slist_reverse (*children);
-
+  folder_unix->asof = time (NULL);
   return TRUE;
 }
 
-static GtkFileInfo *
-filename_get_info (const gchar     *filename,
-		   GtkFileInfoType  types,
-		   GError         **error)
+static gboolean
+cb_fill_in_stats (gpointer key, gpointer value, gpointer user_data)
 {
-  GtkFileInfo *info;
-  struct stat statbuf;
-  gboolean do_stat = (types & (GTK_FILE_INFO_IS_FOLDER |
-			       GTK_FILE_INFO_IS_HIDDEN |
-			       GTK_FILE_INFO_MODIFICATION_TIME |
-			       GTK_FILE_INFO_SIZE));
+  const char *basename = key;
+  struct stat_info_entry *entry = value;
+  GtkFileFolderUnix *folder_unix = user_data;
+  char *fullname = g_build_filename (folder_unix->filename, basename, NULL);
+  gboolean result;
 
-  /* If stat fails, try to fall back to lstat to catch broken links
-   */
-  if (do_stat && stat (filename, &statbuf) != 0)
-    {
-      if (errno != ENOENT || lstat (filename, &statbuf) != 0)
-	{
-	  int save_errno = errno;
-	  gchar *filename_utf8 = g_filename_to_utf8 (filename, -1, NULL, NULL, NULL);
-	  g_set_error (error,
-		       GTK_FILE_SYSTEM_ERROR,
-		       GTK_FILE_SYSTEM_ERROR_NONEXISTENT,
-		       _("error getting information for '%s': %s"),
-		       filename_utf8 ? filename_utf8 : "???",
-		       g_strerror (save_errno));
-	  g_free (filename_utf8);
-
-	  return NULL;
-	}
-    }
-
-  info = gtk_file_info_new ();
-
-  if (filename_is_root (filename))
-    {
-      if (types & GTK_FILE_INFO_DISPLAY_NAME)
-	gtk_file_info_set_display_name (info, "/");
-
-      if (types & GTK_FILE_INFO_IS_HIDDEN)
-	gtk_file_info_set_is_hidden (info, FALSE);
-    }
+  if (stat (fullname, &entry->statbuf) == -1 &&
+      (errno != ENOENT || lstat (fullname, &entry->statbuf) == -1))
+    result = TRUE;  /* Couldn't stat -- remove from hash.  */
   else
-    {
-      gchar *basename = g_path_get_basename (filename);
+    result = FALSE;
 
-      if (types & GTK_FILE_INFO_DISPLAY_NAME)
-	{
-	  gchar *display_name = g_filename_to_utf8 (basename, -1, NULL, NULL, NULL);
-	  if (!display_name)
-	    display_name = g_strescape (basename, NULL);
+  g_free (fullname);
+  return result;
+}
 
-	  gtk_file_info_set_display_name (info, display_name);
 
-	  g_free (display_name);
-	}
+static gboolean
+fill_in_stats (GtkFileFolderUnix *folder_unix, GError **error)
+{
+  if (folder_unix->have_stat)
+    return TRUE;
 
-      if (types & GTK_FILE_INFO_IS_HIDDEN)
-	{
-	  gtk_file_info_set_is_hidden (info, basename[0] == '.');
-	}
+  if (!fill_in_names (folder_unix, error))
+    return FALSE;
 
-      g_free (basename);
-    }
+#if 0
+  g_print ("Stating directory %s\n", folder_unix->filename);
+#endif
+  g_hash_table_foreach_remove (folder_unix->stat_info,
+			       cb_fill_in_stats,
+			       folder_unix);
 
-  if (types & GTK_FILE_INFO_IS_FOLDER)
-    {
-      gtk_file_info_set_is_folder (info, S_ISDIR (statbuf.st_mode));
-   }
+  folder_unix->have_stat = TRUE;
+  return TRUE;
+}
 
-  if (types & GTK_FILE_INFO_MIME_TYPE)
-    {
-      const char *mime_type = xdg_mime_get_mime_type_for_file (filename);
-      gtk_file_info_set_mime_type (info, mime_type);
-    }
 
-  if (types & GTK_FILE_INFO_MODIFICATION_TIME)
-    {
-      gtk_file_info_set_modification_time (info, statbuf.st_mtime);
-    }
+static gboolean
+cb_fill_in_mime_type (gpointer key, gpointer value, gpointer user_data)
+{
+  const char *basename = key;
+  struct stat_info_entry *entry = value;
+  GtkFileFolderUnix *folder_unix = user_data;
+  char *fullname = g_build_filename (folder_unix->filename, basename, NULL);
 
-  if (types & GTK_FILE_INFO_SIZE)
-    {
-      gtk_file_info_set_size (info, (gint64)statbuf.st_size);
-    }
+  /* FIXME: Should not need to re-stat.  */
+  const char *mime_type = xdg_mime_get_mime_type_for_file (fullname);
+  entry->mime_type = g_strdup (mime_type);
 
-  return info;
+  g_free (fullname);
+  /* FIXME: free on NULL?  */
+  return FALSE;
+}
+
+static gboolean
+fill_in_mime_type (GtkFileFolderUnix *folder_unix, GError **error)
+{
+  if (folder_unix->have_mime_type)
+    return TRUE;
+
+#if 0
+  g_print ("Getting mime types for directory %s\n", folder_unix->filename);
+#endif
+  g_hash_table_foreach_remove (folder_unix->stat_info,
+			       cb_fill_in_mime_type,
+			       folder_unix);
+
+  folder_unix->have_mime_type = TRUE;
+  return TRUE;
 }
 
 static GtkFilePath *
