@@ -18,10 +18,10 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
 #include "gtkcontainer.h"
 #include "gtkintl.h"
 #include "gtkprivate.h"
-#include "gtksignal.h"
 #include "gtksizegroup.h"
 
 enum {
@@ -48,7 +48,7 @@ static void add_widget_to_closure (GtkWidget         *widget,
 				   GSList           **widgets);
 
 static GQuark size_groups_quark;
-static const gchar *size_groups_tag = "gtk-size-groups";
+static const gchar size_groups_tag[] = "gtk-size-groups";
 
 static GSList *
 get_size_groups (GtkWidget *widget)
@@ -239,9 +239,9 @@ gtk_size_group_class_init (GtkSizeGroupClass *klass)
   g_object_class_install_property (gobject_class,
 				   PROP_MODE,
 				   g_param_spec_enum ("mode",
-						      _("Mode"),
-						      _("The directions in which the size group effects the requested sizes"
-							" of its component widgets."),
+						      P_("Mode"),
+						      P_("The directions in which the size group effects the requested sizes"
+							" of its component widgets"),
 						      GTK_TYPE_SIZE_GROUP_MODE,
 						      GTK_SIZE_GROUP_HORIZONTAL,
 						      G_PARAM_READWRITE));
@@ -256,10 +256,10 @@ gtk_size_group_init (GtkSizeGroup *size_group)
   size_group->have_height = 0;
 }
 
-GtkType
+GType
 gtk_size_group_get_type (void)
 {
-  static GtkType size_group_type = 0;
+  static GType size_group_type = 0;
 
   if (!size_group_type)
     {
@@ -276,7 +276,8 @@ gtk_size_group_get_type (void)
 	(GInstanceInitFunc) gtk_size_group_init,
       };
 
-      size_group_type = g_type_register_static (G_TYPE_OBJECT, "GtkSizeGroup", &size_group_info, 0);
+      size_group_type = g_type_register_static (G_TYPE_OBJECT, "GtkSizeGroup",
+						&size_group_info, 0);
     }
 
   return size_group_type;
@@ -420,10 +421,11 @@ gtk_size_group_add_widget (GtkSizeGroup     *size_group,
 
       size_group->widgets = g_slist_prepend (size_group->widgets, widget);
 
-      gtk_signal_connect (GTK_OBJECT (widget), "destroy",
-			  GTK_SIGNAL_FUNC (gtk_size_group_widget_destroyed), size_group);
+      g_signal_connect (widget, "destroy",
+			G_CALLBACK (gtk_size_group_widget_destroyed),
+			size_group);
 
-      g_object_ref (G_OBJECT (size_group));
+      g_object_ref (size_group);
     }
   
   queue_resize_on_group (size_group);
@@ -446,8 +448,9 @@ gtk_size_group_remove_widget (GtkSizeGroup     *size_group,
   g_return_if_fail (GTK_IS_WIDGET (widget));
   g_return_if_fail (g_slist_find (size_group->widgets, widget));
 
-  gtk_signal_disconnect_by_func (GTK_OBJECT (widget), 
-				 GTK_SIGNAL_FUNC (gtk_size_group_widget_destroyed), size_group);
+  g_signal_handlers_disconnect_by_func (widget,
+					gtk_size_group_widget_destroyed,
+					size_group);
   
   groups = get_size_groups (widget);
   groups = g_slist_remove (groups, size_group);
@@ -457,7 +460,7 @@ gtk_size_group_remove_widget (GtkSizeGroup     *size_group,
   queue_resize_on_group (size_group);
   gtk_widget_queue_resize (widget);
 
-  g_object_unref (G_OBJECT (size_group));
+  g_object_unref (size_group);
 }
 
 static gint
@@ -488,7 +491,9 @@ do_size_request (GtkWidget *widget)
   if (GTK_WIDGET_REQUEST_NEEDED (widget))
     {
       gtk_widget_ensure_style (widget);
-      gtk_signal_emit_by_name (GTK_OBJECT (widget), "size_request", &widget->requisition);
+      g_signal_emit_by_name (widget,
+			     "size_request",
+			     &widget->requisition);
       
       GTK_PRIVATE_UNSET_FLAG (widget, GTK_REQUEST_NEEDED);
     }
