@@ -18,8 +18,10 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
 #include "gtktreemodelfilter.h"
 #include "gtkintl.h"
+#include "gtktreednd.h"
 #include <string.h>
 
 /* ITER FORMAT:
@@ -110,6 +112,7 @@ enum
 static void         gtk_tree_model_filter_init                            (GtkTreeModelFilter      *filter);
 static void         gtk_tree_model_filter_class_init                      (GtkTreeModelFilterClass *filter_class);
 static void         gtk_tree_model_filter_tree_model_init                 (GtkTreeModelIface       *iface);
+static void         gtk_tree_model_filter_drag_source_init                (GtkTreeDragSourceIface  *iface);
 static void         gtk_tree_model_filter_finalize                        (GObject                 *object);
 static void         gtk_tree_model_filter_set_property                    (GObject                 *object,
                                                                            guint                    prop_id,
@@ -177,6 +180,14 @@ static void         gtk_tree_model_filter_ref_node                        (GtkTr
 static void         gtk_tree_model_filter_unref_node                      (GtkTreeModel           *model,
                                                                            GtkTreeIter            *iter);
 
+/* TreeDragSource interface */
+static gboolean    gtk_tree_model_filter_row_draggable                    (GtkTreeDragSource      *drag_source,
+                                                                           GtkTreePath            *path);
+static gboolean    gtk_tree_model_filter_drag_data_get                    (GtkTreeDragSource      *drag_source,
+                                                                           GtkTreePath            *path,
+                                                                           GtkSelectionData       *selection_data);
+static gboolean    gtk_tree_model_filter_drag_data_delete                 (GtkTreeDragSource      *drag_source,
+                                                                           GtkTreePath            *path);
 
 /* private functions */
 static void        gtk_tree_model_filter_build_level                      (GtkTreeModelFilter     *filter,
@@ -259,6 +270,13 @@ gtk_tree_model_filter_get_type (void)
           NULL
         };
 
+      static const GInterfaceInfo drag_source_info =
+        {
+          (GInterfaceInitFunc) gtk_tree_model_filter_drag_source_init,
+          NULL,
+          NULL
+        };
+
       tree_model_filter_type = g_type_register_static (G_TYPE_OBJECT,
                                                        "GtkTreeModelFilter",
                                                        &tree_model_filter_info, 0);
@@ -266,6 +284,10 @@ gtk_tree_model_filter_get_type (void)
       g_type_add_interface_static (tree_model_filter_type,
                                    GTK_TYPE_TREE_MODEL,
                                    &tree_model_info);
+
+      g_type_add_interface_static (tree_model_filter_type,
+                                   GTK_TYPE_TREE_DRAG_SOURCE,
+                                   &drag_source_info);
     }
 
   return tree_model_filter_type;
@@ -336,6 +358,14 @@ gtk_tree_model_filter_tree_model_init (GtkTreeModelIface *iface)
   iface->unref_node = gtk_tree_model_filter_unref_node;
 }
 
+static void
+gtk_tree_model_filter_drag_source_init (GtkTreeDragSourceIface *iface)
+{
+  iface->row_draggable = gtk_tree_model_filter_row_draggable;
+  iface->drag_data_delete = gtk_tree_model_filter_drag_data_delete;
+  iface->drag_data_get = gtk_tree_model_filter_drag_data_get;
+}
+
 
 static void
 gtk_tree_model_filter_finalize (GObject *object)
@@ -963,7 +993,7 @@ bsearch_elt_with_offset (GArray *array,
         return NULL;
     }
 
-  while (start != end)
+  do
     {
       middle = (start + end) / 2;
 
@@ -976,6 +1006,7 @@ bsearch_elt_with_offset (GArray *array,
       else
         break;
     }
+  while (start != end);
 
   if (elt->offset == offset)
     {
@@ -1050,13 +1081,14 @@ gtk_tree_model_filter_row_changed (GtkTreeModel *c_model,
     {
       /* get rid of this node */
       gtk_tree_model_get_iter (GTK_TREE_MODEL (filter), &iter, path);
-      gtk_tree_model_filter_remove_node (filter, &iter, TRUE);
 
       level = FILTER_LEVEL (iter.user_data);
 
       if (!level->parent_level)
         filter->priv->root_level_visible--;
 
+      gtk_tree_model_filter_remove_node (filter, &iter, TRUE);
+
       goto done;
     }
 
@@ -1139,8 +1171,8 @@ gtk_tree_model_filter_row_inserted (GtkTreeModel *c_model,
                                     gpointer      data)
 {
   GtkTreeModelFilter *filter = GTK_TREE_MODEL_FILTER (data);
-  GtkTreePath *path;
-  GtkTreePath *real_path;
+  GtkTreePath *path = NULL;
+  GtkTreePath *real_path = NULL;
   GtkTreeIter iter;
 
   GtkTreeIter real_c_iter;
@@ -1319,14 +1351,19 @@ done_and_emit:
                                                                 FALSE, TRUE);
 
   if (!path)
-    return;
+    goto done;
 
   gtk_tree_model_filter_increment_stamp (filter);
 
   gtk_tree_model_get_iter (GTK_TREE_MODEL (data), &iter, path);
   gtk_tree_model_row_inserted (GTK_TREE_MODEL (data), path, &iter);
 
+  gtk_tree_path_free (path);
+
 done:
+  if (real_path)
+    gtk_tree_path_free (real_path);
+
   if (free_c_path)
     gtk_tree_path_free (c_path);
 }
@@ -1430,88 +1467,79 @@ gtk_tree_model_filter_row_deleted (GtkTreeModel *c_model,
                                                                 c_path,
                                                                 FALSE,
                                                                 FALSE);
+
   if (!path)
     {
-      path = gtk_real_tree_model_filter_convert_child_path_to_path (filter,
-                                                                    c_path,
-                                                                    FALSE,
-                                                                    TRUE);
+      /* fixup the offsets */
+      GtkTreePath *real_path;
 
-      if (!path)
+      if (!filter->priv->root)
+        return;
+
+      level = FILTER_LEVEL (filter->priv->root);
+
+      /* subtract vroot if necessary */
+      if (filter->priv->virtual_root)
         {
-          /* fixup the offsets */
-          GtkTreePath *real_path;
-
-          if (!filter->priv->root)
+          real_path = gtk_tree_model_filter_remove_root (c_path,
+                                                         filter->priv->virtual_root);
+          /* we don't handle this */
+          if (!real_path)
             return;
+        }
+      else
+        real_path = gtk_tree_path_copy (c_path);
 
-          level = FILTER_LEVEL (filter->priv->root);
-
-          /* subtract vroot if necessary */
-          if (filter->priv->virtual_root)
+      i = 0;
+      if (gtk_tree_path_get_depth (real_path) - 1 >= 1)
+        {
+          while (i < gtk_tree_path_get_depth (real_path) - 1)
             {
-              real_path = gtk_tree_model_filter_remove_root (c_path,
-                                                             filter->priv->virtual_root);
-              /* we don't handle this */
-              if (!real_path)
-                return;
-            }
-          else
-            real_path = gtk_tree_path_copy (c_path);
+              gint j;
 
-          i = 0;
-          if (gtk_tree_path_get_depth (real_path) - 1 >= 1)
-            {
-              while (i < gtk_tree_path_get_depth (real_path) - 1)
+              if (!level)
                 {
-                  gint j;
-
-                  if (!level)
-                    {
-                      /* we don't cover this */
-                      gtk_tree_path_free (real_path);
-                      return;
-                    }
-
-                  elt = bsearch_elt_with_offset (level->array,
-                                                 gtk_tree_path_get_indices (real_path)[i],
-                                                 &j);
-
-                  if (!elt || !elt->children)
-                    {
-                      /* parent is filtered out, so no level */
-                      gtk_tree_path_free (real_path);
-                      return;
-                    }
-
-                  level = elt->children;
-                  i++;
+                  /* we don't cover this */
+                  gtk_tree_path_free (real_path);
+                  return;
                 }
+
+              elt = bsearch_elt_with_offset (level->array,
+                                             gtk_tree_path_get_indices (real_path)[i],
+                                             &j);
+
+              if (!elt || !elt->children)
+                {
+                  /* parent is filtered out, so no level */
+                  gtk_tree_path_free (real_path);
+                  return;
+                }
+
+              level = elt->children;
+              i++;
             }
-
-          offset = gtk_tree_path_get_indices (real_path)[gtk_tree_path_get_depth (real_path) - 1];
-          gtk_tree_path_free (real_path);
-
-          if (!level)
-            return;
-
-          /* we need:
-           * - the offset of the removed item
-           * - the level
-           */
-          for (i = 0; i < level->array->len; i++)
-            {
-              elt = &g_array_index (level->array, FilterElt, i);
-              if (elt->offset > offset)
-                elt->offset--;
-              if (elt->children)
-                elt->children->parent_elt = elt;
-            }
-
-          return;
         }
 
-      emit_signal = FALSE;
+      offset = gtk_tree_path_get_indices (real_path)[gtk_tree_path_get_depth (real_path) - 1];
+      gtk_tree_path_free (real_path);
+
+      if (!level)
+        return;
+
+      /* we need:
+       * - the offset of the removed item
+       * - the level
+       */
+      for (i = 0; i < level->array->len; i++)
+        {
+          elt = &g_array_index (level->array, FilterElt, i);
+          if (elt->offset > offset)
+            elt->offset--;
+          if (elt->children)
+            elt->children->parent_elt = elt;
+        }
+
+      return;
     }
 
   gtk_tree_model_get_iter (GTK_TREE_MODEL (data), &iter, path);
@@ -1694,7 +1722,10 @@ gtk_tree_model_filter_rows_reordered (GtkTreeModel *c_model,
     }
 
   if (level->array->len < 1)
-    return;
+    {
+      gtk_tree_path_free (path);
+      return;
+    }
 
   /* NOTE: we do not bail out here if level->array->len < 2 like
    * GtkTreeModelSort does. This because we do some special tricky
@@ -1756,7 +1787,15 @@ gtk_tree_model_filter_rows_reordered (GtkTreeModel *c_model,
 static guint
 gtk_tree_model_filter_get_flags (GtkTreeModel *model)
 {
+  GtkTreeModelFlags flags;
+
   g_return_val_if_fail (GTK_IS_TREE_MODEL_FILTER (model), 0);
+  g_return_val_if_fail (GTK_TREE_MODEL_FILTER (model)->priv->child_model != NULL, 0);
+
+  flags = gtk_tree_model_get_flags (GTK_TREE_MODEL_FILTER (model)->priv->child_model);
+
+  if ((flags & GTK_TREE_MODEL_LIST_ONLY) == GTK_TREE_MODEL_LIST_ONLY)
+    return GTK_TREE_MODEL_LIST_ONLY;
 
   return 0;
 }
@@ -2226,6 +2265,62 @@ gtk_tree_model_filter_real_unref_node (GtkTreeModel *model,
     }
 }
 
+/* TreeDragSource interface implementation */
+static gboolean
+gtk_tree_model_filter_row_draggable (GtkTreeDragSource *drag_source,
+                                     GtkTreePath       *path)
+{
+  GtkTreeModelFilter *tree_model_filter = (GtkTreeModelFilter *)drag_source;
+  GtkTreePath *child_path;
+  gboolean draggable;
+
+  g_return_val_if_fail (GTK_IS_TREE_MODEL_FILTER (drag_source), FALSE);
+  g_return_val_if_fail (path != NULL, FALSE);
+
+  child_path = gtk_tree_model_filter_convert_path_to_child_path (tree_model_filter, path);
+  draggable = gtk_tree_drag_source_row_draggable (GTK_TREE_DRAG_SOURCE (tree_model_filter->priv->child_model), child_path);
+  gtk_tree_path_free (child_path);
+
+  return draggable;
+}
+
+static gboolean
+gtk_tree_model_filter_drag_data_get (GtkTreeDragSource *drag_source,
+                                     GtkTreePath       *path,
+                                     GtkSelectionData  *selection_data)
+{
+  GtkTreeModelFilter *tree_model_filter = (GtkTreeModelFilter *)drag_source;
+  GtkTreePath *child_path;
+  gboolean gotten;
+
+  g_return_val_if_fail (GTK_IS_TREE_MODEL_FILTER (drag_source), FALSE);
+  g_return_val_if_fail (path != NULL, FALSE);
+
+  child_path = gtk_tree_model_filter_convert_path_to_child_path (tree_model_filter, path);
+  gotten = gtk_tree_drag_source_drag_data_get (GTK_TREE_DRAG_SOURCE (tree_model_filter->priv->child_model), child_path, selection_data);
+  gtk_tree_path_free (child_path);
+
+  return gotten;
+}
+
+static gboolean
+gtk_tree_model_filter_drag_data_delete (GtkTreeDragSource *drag_source,
+                                        GtkTreePath       *path)
+{
+  GtkTreeModelFilter *tree_model_filter = (GtkTreeModelFilter *)drag_source;
+  GtkTreePath *child_path;
+  gboolean deleted;
+
+  g_return_val_if_fail (GTK_IS_TREE_MODEL_FILTER (drag_source), FALSE);
+  g_return_val_if_fail (path != NULL, FALSE);
+
+  child_path = gtk_tree_model_filter_convert_path_to_child_path (tree_model_filter, path);
+  deleted = gtk_tree_drag_source_drag_data_delete (GTK_TREE_DRAG_SOURCE (tree_model_filter->priv->child_model), child_path);
+  gtk_tree_path_free (child_path);
+
+  return deleted;
+}
+
 /* bits and pieces */
 static void
 gtk_tree_model_filter_set_model (GtkTreeModelFilter *filter,
@@ -2391,12 +2486,17 @@ gtk_tree_model_filter_set_visible_func (GtkTreeModelFilter            *filter,
  * @filter: A #GtkTreeModelFilter.
  * @n_columns: The number of columns in the filter model.
  * @types: The #GType<!-- -->s of the columns.
- * @func: A #GtkTreeModelFilterModifyFunc, or %NULL.
+ * @func: A #GtkTreeModelFilterModifyFunc
  * @data: User data to pass to the modify function, or %NULL.
  * @destroy: Destroy notifier of @data, or %NULL.
  *
- * Sets the @filter to have @n_columns columns with @types. If @func
- * is not %NULL, it will set @func to be the modify function of @filter.
+ * With the @n_columns and @types parameters, you give an array of column
+ * types for this model (which will be exposed to the parent model/view).
+ * The @func, @data and @destroy parameters are for specifying the modify
+ * function. The modify function will get called for <emphasis>each</emphasis>
+ * data access, the goal of the modify function is to return the data which 
+ * should be displayed at the location specified using the parameters of the 
+ * modify function.
  *
  * Since: 2.4
  */
